%{
  /* basico03.l
   * Tiny BASIC no GOTO lexer
   * Crenshaw, TINY: use char instead of const for symbols
   *
   * by Andre Adrian
   *
   * version 11jun2006
   */ 
  #include "y.tab.h"
  #include <stdlib.h>
  void yyerror(char *);
  
  extern int namespace; /* import from symtab.c */
  extern char ident[];  /* import from codegen.c */
%}

%%

^[ \t]*\n     	  ;   /* skip empty line */
^[ \t]*"//".*\n   ;   /* skip pure one-line comment */
[ \t]             ;   /* skip whitespace */
\\[ \t]*\n    	  ;   /* skip line continuation */

[-+*/%&|^~,:;()=#<>]  { return *yytext; }
"["                   { return *yytext; }
"]"                   { return *yytext; }

'\\n'       { ident[0] = 10; return 'c'; }
'\\t'       { ident[0] = 9;  return 'c'; }
'.'         { ident[0] = yytext[1];
              return 'c';   /* char const */
            }
            
\".+\"      { strncpy(ident, yytext, 127);
              return 's';   /* array char const */
            }
            
[0-9]+      { strncpy(ident, yytext, 127);
              return 'n';   /* int const */
            }

"array"     { return 'a'; }
"begin"     { return 'b'; }
"break"     { return 'B'; }
"char"      { return 'C'; }
"do"	    { return 'd'; }
"end"       { return 'e'; }
"endif"	    { return 'E'; }
"forward"   { return 'f'; }
"func"      { return 'F'; }
"if"  	    { return 'i'; }
"int"       { return 'I'; }
"else"	    { return 'l'; }
"return"    { return 'r'; }
"then"	    { return 't'; }
"var"       { return 'v'; }
"void"      { return 'V'; }
"wend"	    { return 'w'; }
"while"	    { return 'W'; }

[A-Za-z][A-Za-z0-9]*  { strncpy(ident, yytext, 31);
                        return 'x';   /* identifier */  
                      }                                 

"//".*\n    { return ';'; }   /* change comment after statement to ; */
\n          { return ';'; }   /* change linefeed after statement to ; */

.           yyerror("Unknown character");

%%

int yywrap(void) {
    return 1;
}
