/* symtab.c 
 * Simple symbol tables
 *
 * The symbol tables are 1-dim char arrays. The symbols
 * are space terminated.
 * symbol tables are
 * keywords for reserved words
 * globals for global vars
 * locals for for formal parameters and local vars
 *
 * Search first keywords, then locals, then globals
 *
 * by Andre Adrian
 *
 * version 10jun2006
 */

#include <string.h>
#include <stdio.h>

// #define MODULTEST 

#define SYMTABLEN 1000
#define SYMBOLS 100

char globals[SYMTABLEN];
char globaltype[SYMBOLS];
char locals[SYMTABLEN];
char localtype[SYMBOLS];
int  localoffset[SYMBOLS];
 
void error(char msg[]) {
  fprintf(stderr, msg);
  fprintf(stderr, "\n");
  exit(1);
}
  
 
int lookup(char symtab[], char name[]) {
// look up name in symbol table
// If found return symbol id else -1
  char cn, cs;
  int in, is;
  int found, sym;

  in = 0;  cn = name[in];
  is = 0;  cs = symtab[is];
  found = 1; sym = 0;
  while (cs != '\0') {	// section end
    if (cn != cs) {
      found = 0;
    }
    is = is + 1; cs = symtab[is];
    if (cn != '\0') {
      in = in + 1; cn = name[in];
    }
    if (cs == ' ') {                    // symbol end
      if (found == 1 && cn == '\0') {
      	return sym;
      } else {
	in = 0;  cn = name[in];
	is = is + 1; cs = symtab[is];
	sym = sym + 1;
	found = 1;
      }
    }
  }
  return -1;
}

int gettype(int offset[], char name[]) {
  int id;
  id = lookup(locals, name);
  if (id >= 0) {
    offset[0] = localoffset[id];
    return localtype[id];
  }
  id = lookup(globals, name);
  if (-1 == id) error("variable not found");
  return globaltype[id];
}

void addsym(char symtab[], char name[]) {
  if (strlen(symtab) + strlen(name) + 2 <= SYMTABLEN) {
    strcat(symtab, name);
    strcat(symtab, " ");
  } else {
    error("symtab full");
  }
}

int newname(char symtab[], char name[]) {
  int ident;
  
  ident = lookup(symtab, name);
  if (ident >= 0) {
    return ident;
  } else {
    addsym(symtab, name);
    return lookup(symtab, name);
  }
}

int addglobal(char name[], char type) {
  int id;
  id = newname(globals, name);
  globaltype[id] = type;
  return id;
}

int addlocal(char name[], char type, int offset) {
  int id;
  id = newname(locals, name);
  localtype[id] = type;
  localoffset[id] = offset;
  return id;
}

void localinit() {
  locals[0] = 0;
}

void getname(char symtab[], char name[], int ident) {  
  int i, j;
  
  i = 0;
  while (ident > 0) {
    if (' ' == symtab[i]) {
      ident = ident -1;
    }
    i = i + 1;
  }
  j = 0;
  while(symtab[i] != ' ') {
    name[j] = symtab[i];
    j = j + 1;
    i = i + 1;
  }
  name[j] = '\0';
}

char *gn(char symtab[], int ident) {
  static char name[32];
  
  getname(symtab, name, ident);
  return name;
}
         
void initsymtab() {
  strcpy(globals, "if then else endif while do break wend proc func return int char ");
}

#ifdef MODULTEST

int main(int argc, char *argv[]) {
  int sym;
  initsymtab();
  char name[32];
  
  sym = lookup(globals, "if");  
  printf("lookup if =%d\n", sym);
  getname(globals, name, sym);
  printf("getname =%s\n", name);
  
  sym = lookup(globals, "iff");  
  printf("lookup iff =%d\n", sym);
  
  sym = lookup(globals, "then");    
  printf("lookup then =%d\n", sym);
  getname(globals, name, sym);
  printf("getname =%s\n", name);
  
  sym = lookup(globals, "char");    
  printf("lookup char =%d\n", sym);
  getname(globals, name, sym);
  printf("getname =%s\n", name);
  
  sym = lookup(globals, "xxx");  
  printf("lookup xxx =%d\n", sym);
  sym = lookup(globals, "yyyyyyyyyyy");  
  printf("lookup yyyyyyyyyyy =%d\n", sym);
  
  sym = lookup(globals, "vara");  
  printf("lookup vara =%d\n", sym);
  newname(globals, "vara");
  sym = lookup(globals, "vara");  
  printf("lookup vara =%d\n", sym);
  getname(globals, name, sym);
  printf("getname =%s\n", name);
  
  sym = lookup(globals, "varb");  
  printf("lookup varb =%d\n", sym);
  newname(globals, "varb");
  sym = lookup(globals, "varb");  
  printf("lookup varb =%d\n", sym);
  getname(globals, name, sym);
  printf("getname =%s\n", name);
  
  newname(globals, "vara");
}

#endif
