%{
  /* basico03.y
   * Tiny BASIC no GOTO parser
   *
   * by Andre Adrian
   *
   * version 11jun2006
   */ 
  #include <stdio.h>
  #include "codegen.h"
  int yylex(void);
  void yyerror(char *);

%}

%%

/* based on Pascal BNF */

program:  
   block4  
   | variable_declaration ';' block4
   ;
   
block4:  
   func_declaration ';'
   ;
   
variable_declaration:  
   'v' global_decl  
   | variable_declaration ';' global_decl
   ;
   
global_decl:
   declid ':' 'I'                 { declint(); }                      
   | declid ':' 'a' 'n' 'I'   { declarrint($4);  }         
   | declid ':' 'C'              { declchar(); }                     
   | declid ':' 'a' 'n' 'C'  { declarrchar($4);  }
   ;

declid:
   'x'                          { declid(); }

func_declaration:
   func
   | func_declaration ';' func
   ;

func:
   'F' funcid '(' param_list ')' ':' return_type ';' block_or_forward
   ;

funcid:
   'x'                        { funcid(); }
   ;

param_list:
   param_decl
   | param_list ',' param_decl
   |  	 /* empty set */
   ;

param_decl:
   declid ':' 'I'                { paramint(); }                   
   | declid ':' 'a' 'I'        { paramarrint(); }            
   | declid ':' 'C'             { paramchar(); }                  
   | declid ':' 'a' 'C'       { paramarrchar(); }           
   ;
      
return_type:
   'I'                           { printf("#func returns int\n"); }        
   | 'C'                        { printf("#func returns char\n"); }       
   | 'V'                        { printf("#func returns void\n"); }       
   ;
   
block_or_forward:  
   pblock3                       
   | 'f'                     { printf("#forward decl\n"); }            
   ;
   
pblock3:  
   pblock4  
   | pvariable_declaration ';' pblock4
   ;
   
pblock4:  
   beginl stmt_list 'e'          { funcepilogue(); }
   ;
   
beginl:
   'b'                        { funcprologue(); }
   ;
   
pvariable_declaration:  
   'v' local_decl  
   | pvariable_declaration ';' local_decl
   ;
   
local_decl:
   declid ':' 'I'                 { localint(); }                      
   | declid ':' 'a' 'n' 'I'     { localarrint($4);  }         
   | declid ':' 'C'               { localchar(); }                     
   | declid ':' 'a' 'n' 'C'     { localarrchar($4);  }
   ;

stmt_list:  
   statement ';'
   | stmt_list statement ';'
   ;  

statement:
   variable '=' expr              { assign(); }
   | 'i' expr thenl stmt_list else_or_endif
   | whilel expr dol stmt_list 'w'   { wendjump(); }
   | callid '(' expr_list ')'     { callproc(); }
   | 'r'                          { retrn(); }
   | 'r' expr                     { retrnvalue(); }
   | 'B'                          { breakjump(); }
   |   	 /* empty set */
   ;
   
variable:
   varid                     
   | varid '[' expr ']'           { assignarray(); }   
   ;  

varid:
   'x'                          { varid(); }
    
whilel:
   'W'             	      	  { whilelabel(); }
   ;

dol:
   'd'             	      	  { dojump(); }
   ;

thenl:
   't'                           { thenjump(); }

else_or_endif:
   'E'                          { endiflabel(); }
   | elsel stmt_list 'E'        { endiflabel(); }
   ;
   
elsel:
   'l'                      { elsejump(); }
   ;

callid:
   'x'                     { callid(); }
   ;
   
expr_list:
   expr                      { callarg(); }                    
   | expr_list ',' expr      { callarg(); }                    
   |  	 /* empty set */
   ;
  
expr:
   addexpr
   | expr '=' addexpr       { eq(); }                                      
   | expr '#' addexpr       { ne(); }                                      
   | expr '<' addexpr       { lt(); }                                      
   | expr '<' '=' addexpr   { le(); }    
   | expr '>' addexpr       { gt(); }                                      
   | expr '>' '=' addexpr   { ge(); }    
   ;

addexpr:
   term
   | addexpr '+' term       { add(); }     
   | addexpr '-' term       { sub(); }    
   | addexpr '|' term       { or(); }     
   | addexpr '^' term       { xor(); }     
   ;

term:
   unary_fact
   | term '*' unary_fact     { mul(); }                      
   | term '/' unary_fact     { div(); }                      
   | term '%' unary_fact     { mod(); }                      
   | term '&' unary_fact     { and(); }                      
   ;

unary_fact: 
   '-' unary_fact            { neg(); }                          
   | '~' unary_fact          { not(); }                          
   | fact
   ;

fact:
   'n'      	      	      { pushconst(); }                     
   | 'c'      	              { pushchrconst(); }                     
   | 's'      	              { pushstrconst(); }                     
   | 'x'                      { pushvar(); }                       
   | arrayid '[' expr ']'     { pusharrayvar(); }            
   | '(' expr ')'
   | callid '(' expr_list ')' { callfunc(); }             
   ;
   
arrayid:
   'x'                        { arrayid(); }
   ;   

%%

void yyerror(char *s) {
    fprintf(stderr, "%s\n", s);
}

int main(void) {
    fileprologue();
#if YYDEBUG
    yydebug = 1;
#endif    
    yyparse();
    fileepilogue();
    return 0;
}
